from typing import List, Tuple

from einops import rearrange
import pandas as pd
from sklearn.base import TransformerMixin
from torch.utils.data import Dataset
from torch import Tensor
import torch


class SaintDataset(Dataset):
    """Contains the data that will be processed in batches, and divide them in categorical and continuous

    :param data: Dataframe containing the data to use for the batches
    :param target: Name of the target column
    :param cat_cols: List of indices of categorical columns
    :param target_categorical: If True the target is categorical, so it is a classification problem
    :param con_cols: List of indices of continuous columns
    :param scaler: a scikit learn scaler used to transform the continuous columns
    """
    def __init__(self, data: pd.DataFrame, target: str, cat_cols: List[str],
                 con_cols: List[str], scaler: TransformerMixin, target_categorical: bool):
        self.X_categorical = self._define_tensor_features(data, cat_cols, torch.int64)
        self.X_continuos = self._define_tensor_features(data, con_cols, torch.float32, scaler)
        self.y = self._define_tensor_target(data, target, target_categorical)

    def __len__(self):
        return len(self.y)

    @staticmethod
    def _define_tensor_features(df: pd.DataFrame, cols: List[str], dtype: torch.dtype,
                                transformer: TransformerMixin = None) -> Tensor:
        """It convert a ndarray fo features in a Tensor

        :param df: Contains the data to put insert in the tensor
        :param cols: list of column names used to selected the data
        :param dtype: The type of returned Tensor
        """
        if len(cols) > 0:
            df = df.loc[:, cols].values
            if transformer:
                df = transformer.transform(df)
            return torch.from_numpy(df).to(dtype=dtype)
        else:
            return rearrange(torch.zeros(df.shape[0], dtype=dtype), 'n -> n 1')

    @staticmethod
    def _define_tensor_target(df: pd.DataFrame, target: str, target_categorical: bool) -> Tensor:
        """It return a Tensor containing the values of the target column

        :param df: Dataframe that contains the target
        :param target: name of the target column
        :param target_categorical: True if the target is categorical, otherwise is False
        """
        if target in df.columns:
            if target_categorical:
                y = torch.from_numpy(df[target].values).to(dtype=torch.int64)
            else:
                y = torch.from_numpy(df[target].values).to(dtype=torch.float32)
                y = rearrange(y, 'n -> n 1')
            return y
        else:
            return torch.zeros(df.shape[0], dtype=torch.float32)

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor, Tensor]:
        """It returns two tensors one for the categorical values and another for the continuous values.
        The target is concatenated as last column or of the categorical tensors or the continuous ones
        based on the type of the target

        :param idx: numeric index of the data that we want to process
        """
        return self.X_categorical[idx], self.X_continuos[idx], self.y[idx]



from typing import List

import torch
import torch.nn.functional as f
from einops import rearrange
from torch import nn, einsum, Tensor


class Residual(nn.Module):
    """Define a residual block given a Pytorch Module, used for the skip connection"""
    def __init__(self, fn: nn.Module):
        """
        :param fn: pytorch Module that will be used in order to create the skip connection
        """
        super().__init__()
        self.fn = fn

    def forward(self, x, *args, **kwargs) -> Tensor:
        return self.fn(x, *args, **kwargs) + x


class PreNorm(nn.Module):
    """Apply Layer Normalization before a Module"""
    def __init__(self, dim, fn: nn.Module):
        """
        :param dim: dimension of embedding in input to the module
        :param fn: pytorch Module that will be applied after the normalization layer
        """
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor, *args, **kwargs) -> Tensor:
        return self.fn(self.norm(x), *args, **kwargs)


class GEGLU(nn.Module):
    """Gated GELU, it splits a tensor in two slices based on the last dimension, and then multiply the
       first half and the gelu of the second half
    """
    def forward(self, x: Tensor) -> Tensor:
        x, gates = x.chunk(2, dim=-1)
        return x * f.gelu(gates)


class Attention(nn.Module):
    """Module that implements self attention"""
    def __init__(self, dim: int, heads: int, dim_head: int):
        """
        :param dim: dimension embedding used in input
        :param heads: number of heads for the Multi Head Attention
        :param dim_head: output dimension given an head
        """
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.scale = dim_head ** -0.5

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Linear(inner_dim, dim)

    def forward(self, x: Tensor) -> Tensor:
        # create query, key value from x
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        # split query, key, value for each head adding a new dimension in the tensors
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.heads), (q, k, v))
        # product matrix between q and k transposed
        logits = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
        attn = logits.softmax(dim=-1)
        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        # concatenate result obtained by different heads
        out = rearrange(out, 'b h n d -> b n (h d)', h=self.heads)
        return self.to_out(out)


class RowColTransformer(nn.Module):
    """Module use to define RowColTransformer"""
    def __init__(self, dim: int, nfeats: int, depth: int, heads: int, dim_head: int,
                 ff_dropout: float, scale_dim_internal_col: float,
                 scale_dim_internal_row: float, style: str):
        """
        :param dim: dimension of embedding in input to the module
        :param nfeats: total number of features, needed for the intersample attention
        :param depth: depth of the network, this imply how many times is applied the attention
        :param heads: number of heads for the Multi Head Attention
        :param dim_head: dimension of embedding used in each head
        :param ff_dropout: probability used in the dropout layer of the feed forward layers applied after the attention
        :param scale_dim_internal_col: scale factor input dimension in order to obtain the output dimension of the
        first linear layer in case of style col
        :param scale_dim_internal_row: scale factor input dimension in order to obtain the output dimension of the
        first linear layer in case of style row
        :param style: can be col, row, or colrow
        """
        super().__init__()
        self.layers = nn.ModuleList([nn.ModuleList([]) for _ in range(depth)])
        self.style = style
        for i in range(depth):
            if "col" in self.style:
                self.layers[i].extend(nn.ModuleList([
                    PreNorm(dim, Residual(Attention(dim, heads=heads, dim_head=dim_head))),
                    PreNorm(dim, Residual(SimpleMLP(dim, int(dim * scale_dim_internal_col),
                                                    dim, GEGLU(), dropout=ff_dropout)))
                ]))
            if "row" in self.style:
                self.layers[i].extend(nn.ModuleList([
                    PreNorm(dim * nfeats, Residual(Attention(dim * nfeats, heads=heads, dim_head=dim_head))),
                    PreNorm(dim * nfeats, Residual(SimpleMLP(dim * nfeats,
                                                             int(dim * nfeats * scale_dim_internal_row),
                                                             dim * nfeats, GEGLU(), dropout=ff_dropout))),
                ]))

    @staticmethod
    def forward_col(x: Tensor, attn: nn.Module, ff: nn.Module) -> Tensor:
        x = attn(x)
        x = ff(x)
        return x

    @staticmethod
    def forward_row(x: Tensor, attn: nn.Module, ff: nn.Module) -> Tensor:
        n = x.shape[1]
        # doing this arrange we have a batch of dimension 1 so we make attention between samples
        x = rearrange(x, 'b n d -> 1 b (n d)')
        x = attn(x)
        x = ff(x)
        x = rearrange(x, '1 b (n d) -> b n d', n=n)
        return x

    def forward(self, x: Tensor, x_cont: Tensor) -> Tensor:
        if x_cont is not None:
            x = torch.cat((x, x_cont), dim=1)
        if self.style == 'colrow':
            for attn_col, ff_col, attn_row, ff_row in self.layers:
                x = self.forward_col(x, attn_col, ff_col)
                x = self.forward_row(x, attn_row, ff_row)
        elif self.style == 'row':
            for attn, ff in self.layers:
                x = self.forward_row(x, attn, ff)
        else:
            for attn, ff in self.layers:
                x = self.forward_col(x, attn, ff)
        return x


class SimpleMLP(nn.Module):
    """Module that implements a MLP that consists of 2 linear layers with an activation layer
    and dropout in between them. In case of GEGLU activation function, there is a multiplier that
    allow to divide the tensor in 2 parts
    """
    def __init__(self, dim: int, dim_internal: int, dim_out: int,
                 activation_module: nn.Module = nn.ReLU(), dropout: float = 0.):
        """
        :param dim: dimension of embedding in input to the module
        :param dim_internal: dimension after the first linear layer
        :param dim_out: output dimension
        :param activation_module: module with the activation function to use between the 2 linear layers
        :param dropout: probability used in the dropout layer
        """
        super().__init__()
        mult = 1
        if activation_module._get_name() == 'GEGLU':
            # need to increase size of tensor due to how it's implemented GEGLU
            mult = 2
        self.layers = nn.Sequential(
            nn.Linear(dim, dim_internal * mult),
            activation_module,
            nn.Dropout(dropout),
            nn.Linear(dim_internal, dim_out),
        )

    def forward(self, x: Tensor) -> Tensor:
        if len(x.shape) == 1:
            x = x.view(x.size(0), -1)
        x = self.layers(x)
        return x


class SepMLP(nn.Module):
    """Module that implements a separable MLP, this means that for each feature is used a different SimpleMLP
    """
    def __init__(self, dim: int, dim_out_for_each_feat: List[int], scale_dim_internal: float, dropout: float = .0):
        """
        :param dim: dimension of embedding in input to the module
        :param dim_out_for_each_feat: output dimension for each SimpleMLP
        :param scale_dim_internal: scale factor input dimension in order to obtain the output dimension of the
        first linear layer
        :param dropout: probability used in the dropout layer
        """
        super().__init__()
        self.len_feats = len(dim_out_for_each_feat)
        self.layers = nn.ModuleList([])
        for i in range(self.len_feats):
            self.layers.append(SimpleMLP(dim=dim, dim_internal=int(dim * scale_dim_internal),
                                         dim_out=dim_out_for_each_feat[i],
                                         dropout=dropout))

    def forward(self, x: Tensor) -> List[Tensor]:
        y_pred = list([])
        for i in range(self.len_feats):
            x_i = x[:, i, :]
            pred = self.layers[i](x_i)
            y_pred.append(pred)
        return y_pred



import torch
import numpy as np


def mixup(x: torch.Tensor, random_index: torch.Tensor, lam: float = 1.0) -> torch.Tensor:
    """It apply mixup augmentation, making a weighted average between a tensor
    and some random element of the tensor taking random rows

    :param x: Tensor on which apply the mixup augmentation
    :param random_index: list of indices used to permute the tensor
    :param lam: weight in the linear combination between the original values and the random permutation
    """
    return lam * x + (1 - lam) * x[random_index, :]


def cutmix(x: torch.Tensor, random_index: torch.Tensor, lam: float = 0.1) -> torch.Tensor:
    """Define how apply cutmix to a tensor

    :param x: Tensor on which apply the cutmix augmentation
    :param random_index: list of indices used to permute the tensor
    :param lam: probability values have 0 in a binary random mask, so it means probability original values will
    be updated
    """
    x_binary_mask = torch.from_numpy(np.random.choice(2, size=x.shape, p=[lam, 1 - lam]))
    x_random = x[random_index, :]
    x_noised = x.clone().detach()
    x_noised[x_binary_mask == 0] = x_random[x_binary_mask == 0]
    return x_noised


def get_random_index(x: torch.Tensor) -> torch.Tensor:
    """Given a tensor it compute random indices between 0 and the number of the first dimension

    :param x: Tensor used to get the number of rows
    """
    batch_size = x.size()[0]
    index = torch.randperm(batch_size)
    return index



from dataclasses import dataclass, field
from typing import Optional, Dict
from enum import Enum


class ConstrativeEnum(Enum):
    simsiam = 'simsiam'
    standard = 'standard'


class ProjectionHeadStyleEnum(Enum):
    same = 'same'
    different = 'different'


class AttentionTypeEnum(Enum):
    col = "col"
    row = "row"
    colrow = "colrow"


@dataclass
class OptimizerConfig:
    """Define the parameters for CutMix augmentation"""
    learning_rate: float = 0.0001  #: value used to specify the learning rate
    other_params: Optional[Dict] = field(default_factory=dict)


@dataclass
class CutMixConfig:
    """Define the parameters for CutMix augmentation"""
    lam: float = 0.1  #: probability original values will be updated


@dataclass
class MixUpConfig:
    """Define the parameters for CutMix augmentation"""
    lam: float = 0.1  #: weight used for the linear combination


@dataclass
class ConstrastiveConfig:
    """Define the parameters for Contrastive pretraining task"""
    dropout: float = .0  #: probability dropout in projection head
    constrastive_type: ConstrativeEnum = ConstrativeEnum.simsiam
    projhead_style: ProjectionHeadStyleEnum = ProjectionHeadStyleEnum.different  #: it is used to project embeddings
    nce_temp: float = 0.5  #: temperature used for the logits in case of standard constrastive type
    weight: float = 0.1  #: weight of the loss for this pretraining task


@dataclass
class DenoisingConfig:
    """Define the parameters for Denoising pretraining task"""
    weight_cross_entropy: float = 0.5  #: weight reconstruction loss for categorical features
    weight_mse: float = 0.5  #: weight reconstruction loss for continuous features
    scale_dim_internal_sepmlp: float = 5  # scale factor of the input dimension for the first linear layer
    dropout: float = .0  #: probability dropout in SepMLP


@dataclass
class AugmentationConfig:
    """Define the parameters used for the augmentations"""
    cutmix: Optional[CutMixConfig] = CutMixConfig()
    mixup: Optional[MixUpConfig] = MixUpConfig()


@dataclass
class PreTrainTaskConfig:
    """Define the parameters used for pretraining tasks"""
    contrastive: Optional[ConstrastiveConfig] = ConstrastiveConfig()
    denoising: Optional[DenoisingConfig] = DenoisingConfig()


@dataclass
class PreTrainConfig:
    """Define parameters for the steps used during the pretraining"""
    aug: AugmentationConfig = AugmentationConfig()
    task: PreTrainTaskConfig = PreTrainTaskConfig()
    optimizer: OptimizerConfig = OptimizerConfig()
    epochs: int = 2  #: number of epochs of training phase
    batch_size: int = 256  #: dimension of batches using by dataloaders


@dataclass
class TrainConfig:
    """Define parameters for the steps used during the training"""
    internal_dimension_output_layer: int = 20  #: internal dimension of the MLP that compute the output
    mlpfory_dropout: float = .0  #: probability dropout in the the MLP used for prediction
    epochs: int = 10  #: number of epochs of training phase
    optimizer: OptimizerConfig = OptimizerConfig()
    batch_size: int = 32  #: dimension of batches using by dataloaders


@dataclass
class TransformerConfig:
    depth: int = 3  #: number of attention blocks used in the transformer
    heads: int = 1  #: number of heads used in the transformer
    dropout: float = 0.1  #: probability dropout in the transformer
    attention_type: AttentionTypeEnum = AttentionTypeEnum.col  #: type of attention
    dim_head: int = 64
    scale_dim_internal_col: float = 4  # scale factor of the input dimension in case of attention_type col
    scale_dim_internal_row: float = 4  # scale factor of the input dimension in case of attention_type row


@dataclass
class NetworkConfig:
    """Define the neural network parameters"""
    transformer: TransformerConfig = TransformerConfig()
    num_workers: int = 0  #: number of cores to use
    embedding_size: int = 10  #: dimension of computed embeddings
    internal_dimension_embed_continuous: int = 100  #: internal dimension of the mlp used to project continuous columns
    dropout_embed_continuous: float = .0  #: dropout used to compute embedding continuous features


@dataclass
class SaintConfig:
    """Define all the parameters used in SAINT"""
    network: NetworkConfig = NetworkConfig()
    pretrain: PreTrainConfig = PreTrainConfig()
    train: TrainConfig = TrainConfig()



import os

from setuptools import setup

package_root = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(package_root, "src/lit_saint/version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

setup(
    name='lit-saint',
    version=version,
    license='MIT',
    description='Pytorch Lightning implementation of SAINT Model',
    author='Luca Actis Grosso',
    author_email='lucaactisgrosso@gmail.com',
    url='https://github.com/Actis92/lit-saint.git',
    download_url=f'https://github.com/Actis92/lit-saint.git/archive/{version}.tar.gz',
    keywords=['TABULAR', 'SELF SUPERVISED', 'PYTORCH LIGHTNING'],
    install_requires=['pytorch-lightning>=1.3.0,<2',
                      'torch>=1.4',
                      'einops>=0.3.0',
                      'pandas>=1.0',
                      'scikit-learn>=0.24.2',
                      'hydra-core>=1.1.0'
                      ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
      ]
)


import copy
from typing import List, Tuple, Optional, Callable, Dict

import torch
import torch.nn.functional as f
import torchmetrics
from einops import rearrange
from pytorch_lightning.core import LightningModule
from torch import nn, Tensor
import numpy as np
from torchmetrics import Metric

from lit_saint.config import SaintConfig
from lit_saint.modules import SimpleMLP, RowColTransformer, SepMLP
from lit_saint.augmentations import cutmix, mixup, get_random_index


class Saint(LightningModule):
    """Contains the definition of the network and how to execute the pretraining tasks, and how perform
    the training, validation and test steps

    :param categories: List with the number of unique values for each categorical column
    :param continuous: List of indices with continuous columns
    :param dim_target: if categorical represent number of classes of the target otherwise is 1
    :param config: configuration of the model
    :param metrics: Dictionary containing custom metrics to compute during training loop
    :param metrics_single_class: boolean flag, if True the metrics are computed separately for each class
    :param optimizer: custom optimizer to compute gradient of the network
    :param loss_fn: custom loss function to be optimized
    """
    def __init__(
            self,
            categories: List[int],
            continuous: List[int],
            dim_target: int,
            config: SaintConfig,
            metrics: Dict[str, Metric] = None,
            metrics_single_class: bool = True,
            optimizer: Callable = torch.optim.Adam,
            loss_fn: Callable = None,
    ):
        super().__init__()
        self.save_hyperparameters("categories", "continuous", "dim_target", "config")
        assert all(map(lambda n: n > 0, categories)), 'number of each category must be positive'
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.mc_dropout = False
        self.config = config
        self.pretraining = False
        self.dim_target = dim_target
        self.num_categories = len(categories)
        self.num_unique_categories = sum(categories)
        self.num_continuous = len(continuous) if len(continuous) > 0 else 1
        self.num_columns = self.num_continuous + self.num_categories
        # define offset in order to have unique value for each category
        cat_mask_offset = f.pad(torch.tensor(categories), (1, 0), value=0)
        self.cat_mask_offset = cat_mask_offset.cumsum(dim=-1)[:-1]
        self._define_network_components(categories)
        self.metrics_single_class = metrics_single_class
        self.metrics = metrics if metrics else {}
        self.train_metrics = self._define_metrics(self.metrics, metrics_single_class)
        self.val_metrics = self._define_metrics(self.metrics, metrics_single_class)

    def _define_metrics(self, metrics: Dict[str, Metric], metrics_single_class: bool) -> Dict[str, Metric]:
        """Define custom metrics computed during training loop

        :param metrics: Custom metrics to compute
        :param metrics_single_class: boolean flag, if True the metrics are computed separately for each class
        """
        metrics_step = {}
        if metrics_single_class:
            for key, value in metrics.items():
                for i in range(self.dim_target):
                    metrics_step[f"{key}_{i}"] = copy.deepcopy(value)
        else:
            for key, value in metrics.items():
                metrics_step[key] = copy.deepcopy(value)
        return metrics_step

    def _define_network_components(self, categories: List[int]) -> None:
        """Define the components for the neural network

        :param categories: Number of unique categories for each categorical column
        """
        self._define_embedding_layers()
        self._define_transformer()
        self._define_mlp(categories)
        self._define_projection_head()
        self.mlpfory = SimpleMLP(self.config.network.embedding_size * self.num_columns,
                                 self.config.train.internal_dimension_output_layer, self.dim_target,
                                 dropout=self.config.train.mlpfory_dropout)

    def _define_transformer(self) -> None:
        """Instantiate the type of Transformed that will be used in SAINT"""
        self.transformer = RowColTransformer(
            dim=self.config.network.embedding_size,
            nfeats=self.num_categories + self.num_continuous,
            depth=self.config.network.transformer.depth,
            heads=self.config.network.transformer.heads,
            dim_head=self.config.network.transformer.dim_head,
            ff_dropout=self.config.network.transformer.dropout,
            style=self.config.network.transformer.attention_type.value,
            scale_dim_internal_col=self.config.network.transformer.scale_dim_internal_col,
            scale_dim_internal_row=self.config.network.transformer.scale_dim_internal_row
        )

    def _define_embedding_layers(self) -> None:
        """Instatiate embedding layers"""
        # embed continuos variables using one different MLP for each column
        self.embedding_continuos = nn.ModuleList([SimpleMLP(1, self.config.network.internal_dimension_embed_continuous,
                                                            self.config.network.embedding_size,
                                                            dropout=self.config.network.dropout_embed_continuous)
                                                  for _ in range(self.num_continuous)])
        # embedding layer categorical columns
        self.embedding_categorical = nn.Embedding(self.num_unique_categories, self.config.network.embedding_size)

    def _define_mlp(self, categories: List[int]) -> None:
        """Define MLP used for the Denoising task"""
        self.mlp1 = SepMLP(dim=self.config.network.embedding_size, dim_out_for_each_feat=categories,
                           scale_dim_internal=self.config.pretrain.task.denoising.scale_dim_internal_sepmlp,
                           dropout=self.config.pretrain.task.denoising.dropout)
        self.mlp2 = SepMLP(dim=self.config.network.embedding_size,
                           dim_out_for_each_feat=list(np.ones(self.num_continuous).astype(int)),
                           scale_dim_internal=self.config.pretrain.task.denoising.scale_dim_internal_sepmlp,
                           dropout=self.config.pretrain.task.denoising.dropout)

    def _define_projection_head(self) -> None:
        """Define projection heads for contrastive learning"""
        self.pt_mlp = SimpleMLP(self.config.network.embedding_size * self.num_columns,
                                6 * self.config.network.embedding_size * self.num_columns // 5,
                                self.config.network.embedding_size * self.num_columns // 2,
                                dropout=self.config.pretrain.task.contrastive.dropout)
        self.pt_mlp2 = SimpleMLP(self.config.network.embedding_size * self.num_columns,
                                 6 * self.config.network.embedding_size * self.num_columns // 5,
                                 self.config.network.embedding_size * self.num_columns // 2,
                                 dropout=self.config.pretrain.task.contrastive.dropout)

    def _embed_data(self, x_categ: Tensor, x_cont: Tensor) -> Tuple[Tensor, Tensor]:
        """Converts categorical and continuous values in embeddings

        :param x_categ: contains the values for the categorical features
        :param x_cont: contains the values for the continuous features
        """
        x_categ = x_categ + self.cat_mask_offset.type_as(x_categ)
        x_categ_enc = self.embedding_categorical(x_categ)
        n1, n2 = x_cont.shape
        _, n3 = x_categ.shape
        x_cont_enc = torch.empty(n1, n2, self.config.network.embedding_size).to(x_categ_enc.device)
        for i in range(self.num_continuous):
            x_cont_enc[:, i, :] = self.embedding_continuos[i](x_cont[:, i])

        return x_categ_enc, x_cont_enc

    def _embeddings_contrastive(self, embed_categ: Tensor, embed_cont: Tensor, embed_categ_noised: Tensor,
                                embed_cont_noised: Tensor, projhead_style) -> Tuple[Tensor, Tensor]:
        """Given embeddings original and noised version of the categorical and continuous features are
        transformed using attention, normalized and eventually projected using MLP

        :param embed_categ: embeddings categorical features
        :param embed_cont: embeddings continuous features
        :param embed_categ_noised: embedding categorical features after the augmentation
        :param embed_cont_noised: embedding continuous features after the augmentation
        :param projhead_style: how project the values, can be same if use same MLP or different if use one MLP for
        the original values and another for the augmented ones
        """
        embed_tranformed = self.transformer(embed_categ, embed_cont)
        embed_transformed_noised = self.transformer(embed_categ_noised, embed_cont_noised)
        embed_tranformed = (embed_tranformed / embed_tranformed.norm(dim=-1, keepdim=True)).flatten(1, 2)
        embed_transformed_noised = (embed_transformed_noised /
                                    embed_transformed_noised.norm(dim=-1, keepdim=True)).flatten(1, 2)
        if projhead_style.value == 'different':
            embed_tranformed = self.pt_mlp(embed_tranformed)
            embed_transformed_noised = self.pt_mlp2(embed_transformed_noised)
        elif projhead_style.value == 'same':
            embed_tranformed = self.pt_mlp(embed_tranformed)
            embed_transformed_noised = self.pt_mlp(embed_transformed_noised)
        else:
            print('Not using projection head')
        return embed_tranformed, embed_transformed_noised

    def _pretraining_augmentation(self, x_categ: Tensor, x_cont: Tensor, embed_categ: Tensor,
                                  embed_cont: Tensor) -> Tuple[Tensor, Tensor]:
        """ It applies pretraining task, can be cutmix on the original values, and mixup that is applied
        instead on the embedded values

        :param x_categ: values of categorical features
        :param x_cont: values of continuous features
        :param embed_categ: embeddings categorical features
        :param embed_cont: embeddings continuous features
        """
        if self.config.pretrain.aug.cutmix:
            random_index = get_random_index(x_categ)
            x_categ_noised = cutmix(x_categ, random_index, self.config.pretrain.aug.cutmix.lam)
            x_cont_noised = cutmix(x_cont, random_index, self.config.pretrain.aug.cutmix.lam)
            embed_categ_noised, embed_cont_noised = self._embed_data(x_categ_noised, x_cont_noised)
        else:
            # if not apply cutmix the noised embeddings are equal to the original embeddings
            embed_categ_noised, embed_cont_noised = embed_categ, embed_cont

        if self.config.pretrain.aug.mixup:
            random_index = get_random_index(embed_categ_noised)
            embed_categ_noised = mixup(embed_categ_noised, random_index, lam=self.config.pretrain.aug.mixup.lam)
            embed_cont_noised = mixup(embed_cont_noised, random_index, lam=self.config.pretrain.aug.mixup.lam)
        return embed_categ_noised, embed_cont_noised

    def _pretraining_denoising(self, x_categ: Tensor, x_cont: Tensor, embed_categ_noised: Tensor,
                               embed_cont_noised: Tensor) -> Tensor:
        """Given the embeddings after the application of augmentation task it tries to reconstruct
        the original values, and compute the loss separating the continuous values, using MSE loss,
        from the categorical where is used the cross entropy loss. Then return the combination of the loss
        functions

        :param x_categ: values of categorical features
        :param x_cont: values of continuous features
        :param embed_categ_noised: embeddings categorical features after the augmentation
        :param embed_cont_noised: embeddings continuous features after the augmentation
        """
        embed_transformed_noised = self.transformer(embed_categ_noised, embed_cont_noised)
        output_categorical = self.mlp1(embed_transformed_noised[:, :self.num_categories, :])
        output_continuos = self.mlp2(embed_transformed_noised[:, self.num_categories:, :])
        output_continuos = torch.cat(output_continuos, dim=1)
        loss_continuos_columns = f.mse_loss(output_continuos, x_cont)
        loss_categorical_columns = Tensor([0]).to(embed_transformed_noised.device)
        # for each categorical column we compute the cross_entropy where in x_categ
        # there is the index of the categorical value obtained using the Label Encoding
        for j in range(self.num_categories - 1):
            loss_categorical_columns += f.cross_entropy(output_categorical[j], x_categ[:, j])
        return torch.mul(self.config.pretrain.task.denoising.weight_cross_entropy, loss_categorical_columns) +\
            torch.mul(self.config.pretrain.task.denoising.weight_mse, loss_continuos_columns)

    def _pretraining_contrastive(self, embed_categ: Tensor, embed_cont: Tensor, embed_categ_noised: Tensor,
                                 embed_cont_noised: Tensor) -> Tensor:
        """Using the contrastive learning the objective is that the embeddings of the original values,
        and the embeddings of the augmented ones must be similar

        :param embed_categ: embeddings original categorical features
        :param embed_cont: embeddings original continuous features
        :param embed_categ_noised: embeddings categorical features after the augmentation
        :param embed_cont_noised: embeddings continuous features after the augmentation
        """
        if self.config.pretrain.task.contrastive.constrastive_type.value == 'standard':
            embed_tranformed, embed_transformed_noised = self._embeddings_contrastive(
                embed_categ, embed_cont, embed_categ_noised, embed_cont_noised,
                self.config.pretrain.task.contrastive.projhead_style)
            # @ is the matrix multiplication operator
            logits_1 = embed_tranformed @ embed_transformed_noised.t() / self.config.pretrain.task.contrastive.nce_temp
            logits_2 = embed_transformed_noised @ embed_tranformed.t() / self.config.pretrain.task.contrastive.nce_temp
            # targets it's a list of numbers from zero to size of the tensor, in order to make that
            # z0 and z0' are similar, but z0 is different from the other indexes
            targets = torch.arange(logits_1.size(0))
            loss_1 = f.cross_entropy(logits_1, targets)
            loss_2 = f.cross_entropy(logits_2, targets)
            return self.config.pretrain.task.contrastive.weight * (loss_1 + loss_2) / 2
        elif self.config.pretrain.task.contrastive.constrastive_type.value == 'simsiam':
            # it apply the concept of simsiam we want that the embedding minimize the cosine similarity
            # the idea is that we want on the diagonal all 1, it means they are equal /because normalized)
            embed_tranformed, embed_transformed_noised = self._embeddings_contrastive(
                embed_categ, embed_cont, embed_categ_noised, embed_cont_noised,
                self.config.pretrain.task.contrastive.projhead_style)
            return - self.config.pretrain.task.contrastive.weight * \
                f.cosine_similarity(embed_tranformed, embed_transformed_noised).add_(-1).sum()

    def configure_optimizers(self):
        if self.pretraining:
            lr = self.config.pretrain.optimizer.learning_rate
            other_params = self.config.pretrain.optimizer.other_params
        else:
            lr = self.config.train.optimizer.learning_rate
            other_params = self.config.pretrain.optimizer.other_params

        optimizer = self.optimizer(self.parameters(), lr=lr, **other_params)
        return optimizer

    def forward(self, x_categ: Tensor, x_cont: Tensor) -> Tensor:
        x_categ_enc, x_cont_enc = self._embed_data(x_categ, x_cont)
        reps = self.transformer(x_categ_enc, x_cont_enc)
        reps = rearrange(reps, 'b h n -> b (h n)')
        y_outs = self.mlpfory(reps)
        return y_outs

    def pretraining_step(self, x_categ: Tensor, x_cont: Tensor) -> Tensor:
        """Defines all the step that must be sued in the pretraing step and return the computed loss

        :param x_categ: values of categorical features
        :param x_cont: values of continuous features
        """
        embed_categ, embed_cont = self._embed_data(x_categ, x_cont)
        loss = Tensor([0]).to(embed_categ.device)
        embed_categ_noised, embed_cont_noised = self._pretraining_augmentation(x_categ, x_cont,
                                                                               embed_categ, embed_cont)
        loss += self._pretraining_contrastive(embed_categ, embed_cont,
                                              embed_categ_noised, embed_cont_noised)
        if self.config.pretrain.task.denoising:
            loss += self._pretraining_denoising(x_categ, x_cont, embed_categ_noised, embed_cont_noised)
        return loss

    def shared_step(self, batch: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
        """It define the commons step executed during training, validation and test

        :param batch: Contains a batch of data
        """
        x_categ, x_cont, target = batch
        if self.pretraining:
            return self.pretraining_step(x_categ, x_cont), torch.Tensor(), torch.Tensor()
        else:
            y_pred = self(x_categ, x_cont)
            if self.loss_fn:
                return self.loss_fn(y_pred, target), y_pred, target
            else:
                if self.dim_target > 1:
                    loss = self._classification_loss(y_pred, target)
                    return loss, y_pred, target
                else:
                    return self._regression_loss(y_pred, target), y_pred, target

    def training_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int) -> Tensor:
        loss, y_pred, target = self.shared_step(batch)
        self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        if not self.pretraining:
            self.compute_metrics(metrics=self.train_metrics, y_pred=y_pred, target=target)
            if len(self.train_metrics) > 0:
                self.log("train_metrics", self.train_metrics, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return loss

    def compute_metrics(self, metrics: Dict[str, Metric], y_pred: Tensor, target: Tensor) -> None:
        """ Compute custom metrics during training loop

        :param metrics: Metrics to compute
        :param y_pred: predicted value
        :param target: true value
        """
        pred = nn.Softmax(dim=-1)(y_pred) if self.dim_target > 1 else y_pred.detach()
        for key, value in metrics.items():
            if self.metrics_single_class:
                index_class_metric = int(key.split("_")[-1])
                target_single_class = target[target == index_class_metric]
                if target_single_class.shape[0] > 0:
                    value(pred[target == index_class_metric], target[target == index_class_metric])
            else:
                value(pred, target)

    def validation_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int) -> Tensor:
        loss, y_pred, target = self.shared_step(batch)
        self.log("val_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        if not self.pretraining:
            self.compute_metrics(metrics=self.val_metrics, y_pred=y_pred, target=target)
            if len(self.val_metrics) > 0:
                self.log("val_metrics", self.val_metrics, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return loss

    def test_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int) -> Tensor:
        loss, _, _ = self.shared_step(batch)
        self.log("test_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return loss

    def predict_step(self, batch: Tuple[Tensor, Tensor, Tensor], batch_idx: int,
                     dataloader_idx: Optional[int] = None) -> Tensor:
        if self.mc_dropout:
            for m in self.modules():
                if m.__class__.__name__.startswith('Dropout'):
                    m.train()
        x_categ, x_cont, _ = batch
        y_pred = self(x_categ, x_cont)
        if self.dim_target > 1:
            return nn.Softmax(dim=-1)(y_pred)
        else:
            return y_pred

    @staticmethod
    def _classification_loss(y_pred: Tensor, target: Tensor) -> Tensor:
        """Loss function used in case of a classification problem

        :param y_pred: Values predicted
        :param target: Values to predict
        """
        return f.cross_entropy(y_pred, target)

    @staticmethod
    def _regression_loss(y_pred: Tensor, target: Tensor) -> Tensor:
        """Loss function used in case of regression problem

        :param y_pred: Values predicted
        :param target: Values to predict
        """
        return f.mse_loss(y_pred, target)

    def set_pretraining(self, pretraining: bool):
        self.pretraining = pretraining

    def set_mcdropout(self, mc_dropout: bool):
        self.mc_dropout = mc_dropout



from typing import Dict

import numpy
import torch
from pandas import DataFrame

from lit_saint import Saint, SaintDatamodule
from pytorch_lightning.trainer import Trainer


class SaintTrainer:
    """This class simplify how to train and make predictions using Saint

    :param pretrainer: The Trainer that will be used during the pretraining phase
    :param trainer: The Trainer that will be used during the training and prediction phase
    :param train_loader_params: parameters used to configure the DataLoader during the training phase
    :param pretrain_loader_params: parameters used to configure the DataLoader during the pretraining phase
    """
    def __init__(self, pretrainer: Trainer = None, trainer: Trainer = None, train_loader_params: Dict = None,
                 pretrain_loader_params: Dict = None):
        self.pretrainer = pretrainer
        self.pretrain_loader_params = pretrain_loader_params if pretrain_loader_params else {"batch_size": 256}
        self.trainer = trainer
        self.train_loader_params = train_loader_params if train_loader_params else {"batch_size": 256}

    def prefit(self, model: Saint, datamodule: SaintDatamodule) -> None:
        """Function that is used for the pretraining of the model

        :param model: instance of Saint that will be used for the pretraining
        :param datamodule: instance of SaintDataModule that contains the data for the pretraining
        """
        model.set_pretraining(True)
        datamodule.set_pretraining(True)
        datamodule.set_data_loader_params(self.pretrain_loader_params)
        self.pretrainer.fit(model=model, datamodule=datamodule)

    def get_model_from_checkpoint(self, model: Saint, pretraining: bool) -> None:
        """Function that load the best checkpoint and update the model inplace

        :param model: instance of Saint that will be updated by the checkpoint
        :param pretraining: boolean flag if true the checkpoint will be taken from the pretrainer
        """
        trainer = self.pretrainer if pretraining else self.trainer
        checkpoint_callback = [c for c in trainer.callbacks if c.__class__.__name__ == 'ModelCheckpoint']
        if len(checkpoint_callback) == 1 and checkpoint_callback[0].best_model_path != "":
            model.load_from_checkpoint(checkpoint_path=checkpoint_callback[0].best_model_path)

    def fit(self, model: Saint, datamodule: SaintDatamodule, enable_pretraining: bool) -> None:
        """Runs the full optimization routine.

        :param model: instance of Saint that will be used for the training
        :param datamodule: instance of SaintDataModule that contains the data for the training
        :param enable_pretraining: boolean flag, if True the pretraining will be executed otherwise only the training
        """
        if enable_pretraining:
            self.prefit(model=model, datamodule=datamodule)
            self.get_model_from_checkpoint(model, pretraining=True)
        model.set_pretraining(False)
        datamodule.set_pretraining(False)
        datamodule.set_data_loader_params(self.train_loader_params)
        self.trainer.fit(model=model, datamodule=datamodule)
        self.get_model_from_checkpoint(model, pretraining=False)

    def predict(self, model: Saint, datamodule: SaintDatamodule, df: DataFrame,
                mc_dropout_iterations: int = 0) -> numpy.ndarray:
        """Separates from fit to make sure you never run on your predictions set until you want to.
        This will call the model forward function to compute predictions.

        :param model: model used for the prediction
        :param datamodule: the datamodule that will preprocess the data before prediction
        :param df: the data taht will be used for prediction
        :param mc_dropout_iterations: number of iterations used for mcdropout estimation, if 0 it will not use
        mcdropout
        """
        datamodule.set_predict_set(df)
        if mc_dropout_iterations > 0:
            model.set_mcdropout(True)
            mc_predictions = []
            for _ in range(mc_dropout_iterations):
                prediction = torch.cat(self.trainer.predict(model, datamodule=datamodule))
                mc_predictions.append(prediction)
            model.set_mcdropout(False)
            return torch.stack(mc_predictions, axis=2).cpu().numpy()
        prediction = self.trainer.predict(model, datamodule=datamodule)
        return torch.cat(prediction).cpu().numpy()



from typing import Dict

from pytorch_lightning import LightningDataModule
import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader

from lit_saint.dataset import SaintDataset


class SaintDatamodule(LightningDataModule):
    """It preprocess the data, doing LabelEncoding for the categorical values and fitting a StandardScaler
    for the numerical columns on the training set. And it splits the data and defines the dataloaders
    """
    NAN_LABEL = "SAINT_NAN"

    def __init__(self, df: pd.DataFrame, target: str, split_column: str, scaler: TransformerMixin = None):
        """
        :param df: contains the data that will be used by the dataLoaders
        :param target: name of the target column
        :param split_column: name of the column used to split the data
        :param scaler: a scikit learn transformer in order to rescale the continuos variables, if not specified
        it will use the StandardScaler
        """
        super().__init__()
        self.target: str = target
        self.pretraining = False
        self.data_loader_params = {"batch_size": 256}
        self.categorical_columns = []
        self.categorical_dims = []
        self.numerical_columns = []
        self.dim_target = 1
        self.target_nan_index = None
        self.dict_label_encoder = {}
        self.predict_set = None
        self.scaler = scaler if scaler else StandardScaler()
        self.split_column = split_column
        self.prep(df)

    def prep(self, df: pd.DataFrame) -> None:
        """It find the indexes for each categorical and continuous columns, and for each categorical it
            applies Label Encoding in order to convert them in integers and save the number of classes for each
            categorical column

        :param df: contains the data that need to be processed
        """
        df = df.copy()
        col_not_to_use = []
        for col in df.columns:
            if df[col].dtypes.name in ["object", "category"]:
                df = self.prep_categorical_columns(col=col, df=df)
            elif df[col].dtypes.name in ["int64", "float64", "int32", "float32"]:
                df = self.prep_continuous_columns(col=col, df=df)
            else:
                col_not_to_use.append(col)
        if len(self.categorical_columns) == 0:
            self.categorical_dims.append(1)
        print("The following cols will not be used because they have a not supported data type: ", col_not_to_use)
        self._split_data(df=df)
        self.scaler_continuous_columns(df=df)

    def prep_categorical_columns(self, col: str, df: pd.DataFrame) -> pd.DataFrame:
        if df[col].isna().any():  # the columns contains nan
            df[col] = df[col].fillna(self.NAN_LABEL)
        if col != self.split_column:
            l_enc = OrdinalEncoder(handle_unknown='use_encoded_value', unknown_value=df[col].nunique())
            df[col] = l_enc.fit_transform(df[col].values.reshape(-1, 1)).astype(int)
            self.dict_label_encoder[col] = l_enc
            if col == self.target:
                if self.NAN_LABEL in l_enc.categories_[0]:
                    self.dim_target = len(l_enc.categories_[0]) - 1
                    self.target_nan_index = list(l_enc.categories_[0]).index(self.NAN_LABEL)
                else:
                    self.dim_target = len(l_enc.categories_[0])
                    self.target_nan_index = None
            else:
                self.categorical_columns.append(col)
                self.categorical_dims.append(len(l_enc.categories_[0]) + 1)
        return df

    def prep_continuous_columns(self, col: str, df: pd.DataFrame) -> pd.DataFrame:
        if df[col].isna().any():  # the columns contains nan
            df[col] = df[col].fillna(0)
        if col != self.target:
            self.numerical_columns.append(col)
        return df

    def scaler_continuous_columns(self, df: pd.DataFrame) -> None:
        """Fit a StandardScaler for each continuos columns on the training set

        :param df: contains the data that need to be processed
        :param split_column: name of column used to split the data
        """
        df_train = df.loc[df[self.split_column] == "train"].loc[:, self.numerical_columns].values
        if len(self.numerical_columns) > 0:
            self.scaler.fit(df_train)

    def _split_data(self, df: pd.DataFrame) -> None:
        """Split the Dataframe in train, validation and test, and drop the split column

        :param df: contains the data that need to be processed
        """
        self.train = df.loc[df[self.split_column] == "train"].reset_index(drop=True)
        self.validation = df.loc[df[self.split_column] == "validation"].reset_index(drop=True)
        self.test = df.loc[df[self.split_column] == "test"].reset_index(drop=True)
        self.train.drop(self.split_column, axis=1, inplace=True)
        self.validation.drop(self.split_column, axis=1, inplace=True)
        self.test.drop(self.split_column, axis=1, inplace=True)

    def set_predict_set(self, df) -> None:
        """Tranform the categorical columns using the OrdinalEncoders fitted before the training and
        save the dataframe in order to make the predictions

        :param df: The data that will be used to make some predictions
        """
        df = df.copy()
        for col, label_enc in self.dict_label_encoder.items():
            if col != self.target or (col == self.target and col in df.columns):
                if df[col].isna().any():  # the columns contains nan
                    df[col] = df[col].fillna(self.NAN_LABEL)
                df[col] = label_enc.transform(df[col].values.reshape(-1, 1)).astype(int)
        df = df.fillna(0)
        self.predict_set = df

    def _remove_rows_without_labels(self, df) -> pd.DataFrame:
        """Remove rows from a dataframe where the label was NaN

        :param df: the dataframe from which remove rows
        """
        df = df.loc[df[self.target] != self.target_nan_index]
        df[self.target] = df[self.target].apply(lambda x: x if x < self.target_nan_index else x - 1)
        return df

    def _create_dataloader(self, df, is_predict=False) -> DataLoader:
        """ Given a dataframe it return a dataloader and eventually without rows
        that have nan labels if not pretraining

        :param df: the dataframe that will be used inside the DataLoader
        :param is_predict: flag that is true if we are executing a prediction
        """
        if not self.pretraining and self.target_nan_index is not None and not is_predict:
            df = self._remove_rows_without_labels(df)
        dataset = SaintDataset(
            data=df,
            target=self.target,
            cat_cols=self.categorical_columns,
            con_cols=self.numerical_columns,
            scaler=self.scaler,
            target_categorical=self.dim_target > 1
        )
        return DataLoader(
            dataset,
            **self.data_loader_params,
        )

    def train_dataloader(self) -> DataLoader:
        """ Function that loads the train set. """
        return self._create_dataloader(self.train)

    def val_dataloader(self) -> DataLoader:
        """ Function that loads the validation set. """
        return self._create_dataloader(self.validation)

    def test_dataloader(self) -> DataLoader:
        """ Function that loads the validation set. """
        return self._create_dataloader(self.test)

    def predict_dataloader(self) -> DataLoader:
        """ Function that loads the dataset for the prediction. """
        return self._create_dataloader(self.predict_set, True)

    def set_pretraining(self, pretraining: bool) -> None:
        """Function used to set the pretraining flag"""
        self.pretraining = pretraining

    def set_data_loader_params(self, data_loader_params: Dict) -> None:
        """Function used to set the parameters used by the DataLoader"""
        self.data_loader_params = data_loader_params
